#!/usr/bin/python3
# -*- coding: utf-8 -*-

__author__ = 'Équipe pédagogique DIU-EIL Lille'
__date__ = 'juin 2019'
__doc__ = """
Module for encoding and decoding base64 data.

base64 is widely used to transform binary files in ASCII files. More specifically,
it is used by the MIME protocol for sending binary files by emails.

:author: {:s}
:date: {:s}
""".format(__author__, __date__)

import sys
from math import ceil


BASE64_SYMBOLS = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
                  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                  'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
                  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
                  'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
                  'w', 'x', 'y', 'z', '0', '1', '2', '3',
                  '4', '5', '6', '7', '8', '9', '+', '/']
'''
Coding table :
`BASE64_SYMBOLS[k]` = symbol in base64 that corresponds to integer k, which can range from 0 to 63.
'''


LINE_LENGTH = 76
'''
Maximal line length
'''

def many_bytes_to_int(many_bytes):
    '''
    Returns an integer consisting of the representation of all the bytes
    in `many_bytes`

    :param many_bytes: A `bytes` object containing several (>= 0) bytes
    :type: bytes
    :return: An integer whose binary representation is the binary \
             representation of the bytes in `many_bytes`. The most\
             significant byte in the returned integer is the first byte
             in `many_bytes`
    :rtype: int
    :CU: None
    :Examples:

    >>> many_bytes_to_int(bytes([253, 12, 71]))
    16583751
    >>> many_bytes_to_int(bytes([23, 12, 71]))
    1510471
    >>> many_bytes_to_int(bytes([16]))
    16
    >>> many_bytes_to_int([])
    0
    >>> many_bytes_to_int(bytes([129, 126]))
    33150
    '''
    res = 0
    for byte in many_bytes:
        res = res << 8 | byte
    return res

def int_to_sextets(integer, nb_bytes):
    '''
    Convert an integer made from `nb_bytes` bytes to at most 4 sextets

    :param integer: The integer to convert
    :type integer: int
    :param nb_bytes: Number of bytes that were used to represent the integer\
                     `integer`
    :type nb_bytes: int
    :return: A list of at most 4 values, each representing a sextet, converted\
             from the initial integer. The first sextet from the list is\
             made from the 6 most significant bits of the integer, the \
             following sextet is made from the following six bits and so on.
    :rtype: list
    :CU: `nb_bytes` <= 3 and `integer` < 2**(8 * nb_bytes)
    :Examples:

    >>> int_to_sextets(many_bytes_to_int(bytes([253, 12, 71])), 3)
    [63, 16, 49, 7]
    >>> int_to_sextets(many_bytes_to_int(bytes([23, 12, 71])), 3)
    [5, 48, 49, 7]
    >>> int_to_sextets(16, 1)
    [4, 0]
    >>> int_to_sextets(0, 0)
    []
    >>> int_to_sextets(0, 1)
    [0, 0]
    >>>
    # 16 on 2 bytes is represented as 0000 0000 0001 0000
    # which corresponds to three sextets: 000000 000001 000000
    # The two last 0s are added to have complete sextets
    >>> int_to_sextets(16, 2)
    [0, 1, 0]
    >>> int_to_sextets(16, 3)
    [0, 0, 0, 16]
    '''
    assert (nb_bytes <= 3 and (1 << (8 * nb_bytes) > integer)),\
           'int_to_sextets: incoherent parameters'
    mask = 0b111111
    nb_sextets = ceil(nb_bytes * 8 / 6)
    integer <<= (nb_sextets * 6 - nb_bytes * 8)
    return [((integer >> (6 * (nb_sextets - i - 1))) & mask) for i in range(nb_sextets)]

def bytes_to_symbols(data):
    '''
    Takes (at most) three bytes of data in input and returns the corresponding
    bas64 symbols.

    :param data: A list of at most three bytes
    :type list:
    :return: Four base64 symbols (or '=') corresponding to the data given in input
    :rtype: str
    :CU: len(data) <= 3
    :Examples:

    >>> bytes_to_symbols([5])
    'BQ=='
    >>> bytes_to_symbols([4, 163])
    'BKM='
    >>> bytes_to_symbols([28, 89, 101])
    'HFll'
    >>> bytes_to_symbols([])
    ''
    '''
    assert len(data) <= 3, 'bytes_to_symbols: data has to many bytes'
    if len(data) == 0:
        return ''
    sextets = int_to_sextets(many_bytes_to_int(bytes(data)), len(data))
    b64_str = ''.join([BASE64_SYMBOLS[s] for s in sextets])
    return b64_str + ((4 - len(sextets)) * '=')


def  base64_encode(source):
    '''
    Encode a file in base64 and outputs the result on standard output.

    :param source: the source filename
    :type source: str
    '''
    input_stream = open(source, 'rb')
    data = input_stream.read(3)
    nb_printed_symbs = 0
    while len(data) > 0:
        print(bytes_to_symbols(data), end='')
        nb_printed_symbs += 4
        if nb_printed_symbs == LINE_LENGTH:
            print()
            nb_printed_symbs = 0
        data = input_stream.read(3)
    if nb_printed_symbs > 0:
        print()
    input_stream.close()

def decode_base64_symbol(symbol):
    '''
    Convert a base64 symbol in integer.

    :param symbol: the symbol to be converted
    :type symbol: str
    :return: the integer corresponding to the base64 symbol
    :rtype: int
    :CU: the symbol is part of the base64 symbols
    :Examples:

    >>> decode_base64_symbol('A')
    0
    >>> decode_base64_symbol('Z')
    25
    >>> decode_base64_symbol('a')
    26
    >>> decode_base64_symbol('z')
    51
    >>> decode_base64_symbol('0')
    52
    >>> decode_base64_symbol('9')
    61
    >>> decode_base64_symbol('+')
    62
    >>> decode_base64_symbol('/')
    63
    >>> decode_base64_symbol('=')
    Traceback (most recent call last):
    ...
    ValueError: '=' is not in list
    '''
    return BASE64_SYMBOLS.index(symbol)

def symbols_to_sextets(symbols):
    '''
    Convert base64 symbols to sextets

    :param symbols: a string of four base64 symbols (and maybe the = sign)
    :type symbols: str
    :return: a list of one to four sextets whose values correspond to the base64 symbols.
            = characters are not converted.
    :rtype: list
    :CU: len(symbols) == 4
    :Examples:

    >>> symbols_to_sextets('BQ==')
    [1, 16]
    >>> symbols_to_sextets('BKM=')
    [1, 10, 12]
    >>> symbols_to_sextets('HFll')
    [7, 5, 37, 37]
    '''
    assert len(symbols) == 4
    return [decode_base64_symbol(c) for c in symbols if c != '=']

def sextets_to_bytes(sextets):
    '''
    Converts the given sextets to a series of bytes

    :param sextets: A list of up to 4 sextets
    :type sextets: list
    :return: A list of several bytes. The first byte corresponds to the eight \
    most significant bits from the two first sextets, and so on. If a few bits\
    do not fill a complete sextet, they are omitted. For instance, when two \
    sextets are provided (12 bits), only one byte is given on the output.\
    The four least significant bits are discarded.
    :rtype: list
    :CU: len(sextets) <= 4
    :Examples:

    >>> sextets_to_bytes([1, 16])
    [5]
    >>> sextets_to_bytes([])
    []
    >>> sextets_to_bytes([1, 10, 12])
    [4, 163]
    >>> sextets_to_bytes([7, 5, 37, 37])
    [28, 89, 101]
    '''
    nb_sextets = len(sextets)
    if nb_sextets == 0:
        return []
    nb_bytes = (nb_sextets * 6) // 8
    value = sum(byte << (6 * (nb_sextets - i - 1)) for i, byte in enumerate(sextets))
    value >>= nb_sextets * 6 - nb_bytes * 8
    mask = (1 << 8) - 1
    return [(value >> (8*(nb_bytes - i - 1))) & mask for  i in range(nb_bytes)]

def symbols_to_bytes(symbols):
    '''
    Convert base64 symbols to bytes

    :param symbols: a string of four base64 symbols (and maybe the = sign)
    :type symbols: str
    :return: a list of one to 3 bytes whose values correspond to the base64 symbols
    :rtype: list
    :CU: len(symbols) == 4
    :Examples:

    >>> symbols_to_bytes('BQ==')
    [5]
    >>> symbols_to_bytes('BKM=')
    [4, 163]
    >>> symbols_to_bytes('HFll')
    [28, 89, 101]
    '''
    return sextets_to_bytes(symbols_to_sextets(symbols))


def process_base64_line(line):
    '''
    Process a line from a base64 input and writes the conversion on the
    standard output

    :param line: a line of a base64 output
    :type line: str
    :CU: len(line) % 4 == 0 and the line only contains base64 symbols (or possibly = signs)

    There is no example given as sys.stdout.buffer.write is not supported
    by doctests. However you can use the small examples (Codage, Codages,
    Codages.) given in the introductory part of this tutorial.
    '''
    number_blocks = len(line) // 4
    for i in range(number_blocks):
        sys.stdout.buffer.write(bytes(symbols_to_bytes(line[i*4:(i+1)*4])))

def base64_decode(source):
    '''
    Decode a source file encoded in base64 and output the result.

    :param source: the filename of the base64 file to decode
    :type source: str
    '''
    input_stream = open(source, 'r')
    line = input_stream.readline()
    while len(line) > 0:
        process_base64_line(line)
        line = input_stream.readline()
    input_stream.close()


if __name__ == '__main__':
    import doctest
    doctest.testmod()
