import random
import doctest
import time
def creer_grille(horiz, vert):
    '''
    Crée une grille de `horiz` cases horizontalement par `vert` cases
    verticalement. La grille est vide : chaque case contient 0 cellule.

    >>> creer_grille(3, 2)
    [[0, 0, 0], [0, 0, 0]]
    >>> creer_grille(0, 0)
    []
    '''
    
    return [ [ 0 for i in range(horiz) ] for j in range(vert) ]

def hauteur_grille(grille):
    '''
    Renvoie le nombre de cases d'une grille verticalement

    >>> hauteur_grille(creer_grille(3, 2))
    2
    >>> hauteur_grille(creer_grille(0, 0))
    0
    '''
    return len(grille)

def largeur_grille(grille):
    '''
    Renvoie le nombre de cases d'une grille horizontalement

    >>> largeur_grille(creer_grille(3, 2))
    3
    >>> largeur_grille(creer_grille(0, 0))
    0
    '''
    if len(grille) > 0:
        return len(grille[0])
    return 0

def creer_grille_aleatoire(horiz, vert, proba):
    '''
    Crée une grille de `horiz` cases horizontalement par `vert` cases
    verticalement.
    Chaque case contient une cellule avec une probabilité `proba`, 0
    dans le cas contraire.

    >>> creer_grille_aleatoire(3, 2, 0)
    [[0, 0, 0], [0, 0, 0]]
    >>> creer_grille_aleatoire(3, 2, 1)
    [[1, 1, 1], [1, 1, 1]]
    >>> creer_grille_aleatoire(0, 0, 1)
    []
    '''
    grille = creer_grille(horiz, vert)

    for i in range(hauteur_grille(grille)):
        for j in range(largeur_grille(grille)):
            if random.random() < proba:
                grille[i][j] = 1

    return grille

def voisins_case(grille, x, y):
    '''
    Retourne une liste du contenu des cases voisines d'une
    case aux coordonnées `x`, `y`.

    Cette fonction renvoie les voisins ligne par ligne en commençant par 
    la gauche.

    >>> voisins_case([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 1, 1)
    [0, 1, 0, 1, 0, 1, 1, 1]
    >>> voisins_case([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 2, 2)
    [0, 0, 1]
    >>> voisins_case([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 0, 2)
    [1, 0, 1]
    '''
    voisins = []
    largeur = largeur_grille(grille)
    hauteur = hauteur_grille(grille)
    for j in range(-1, 2):
        for i in range(-1, 2):
            if x+i >= 0 and x+i < largeur and y+j >= 0 and y+j < hauteur and\
               (i != 0 or j != 0):
                voisins.append(grille[y+j][x+i])

    return voisins

def nb_cellules_voisins(grille, x, y):
    '''
    Renvoie le nombre de cellules dans les cases voisines d'une case
    aux coordonnées `x`, `y`

    >>> nb_cellules_voisins([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 1, 1)
    5
    >>> nb_cellules_voisins([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 2, 2)
    1
    >>> nb_cellules_voisins([[0, 1, 0], [1, 0, 0], [1, 1, 1]], 0, 2)
    2
    '''

    return sum(voisins_case(grille, x, y))

def copie_grille(grille):
    '''
    Renvoie une copie profonde de la grille passée en paramètre

    >>> grille = [[0, 1, 0], [1, 0, 0], [1, 1, 1]]
    >>> copie_grille(grille)
    [[0, 1, 0], [1, 0, 0], [1, 1, 1]]
    >>> grille2 = copie_grille(grille)
    >>> grille2[0][0] = 1
    >>> grille != grille2 # Nous verifions qu'il s'agit bien de deux grilles différentes 
    True
    '''
    copie = []
    for ligne in grille:
        copie.append(ligne.copy())
    return copie


def afficher_grille(grille):
    '''
    Affiche une grille du jeu de la vie

    >>> afficher_grille([[0, 1, 0], [1, 0, 0], [1, 1, 1]])
    _ O _
    O _ _
    O O O
    >>> afficher_grille(creer_grille(3, 2))
    _ _ _
    _ _ _
    '''

    for line in grille:
        # Crée une liste des caractères à afficher
        chars = [ '_' if n == 0 else 'O' for n in line ]
        print (' '.join(chars))

def generation_suivante(grille):
    '''
    Renvoie une grille correspondant à la génération suivante de la grille
    passée en paramètre.

    >>> generation_suivante([[0, 1, 0], [1, 0, 0], [1, 1, 1]])
    [[0, 0, 0], [1, 0, 1], [1, 1, 0]]
    >>> generation_suivante([[0, 0, 0], [1, 0, 1], [1, 1, 0]])
    [[0, 0, 0], [1, 0, 0], [1, 1, 0]]
    >>> generation_suivante([[0, 0, 0], [1, 0, 0], [1, 1, 0]])
    [[0, 0, 0], [1, 1, 0], [1, 1, 0]]
    '''
    nouvelle = copie_grille(grille)

    for i in range(hauteur_grille(grille)):
        for j in range(largeur_grille(grille)):
            nb_voisins = nb_cellules_voisins(grille, j, i)
            if grille[i][j] == 0 and nb_voisins == 3:
                nouvelle[i][j] = 1
            elif grille[i][j] == 1 and (nb_voisins < 2 or nb_voisins > 3):
                nouvelle[i][j] = 0
    return nouvelle

def evolution_n_generations(grille, n):
    for i in range(n):
        afficher_grille(grille)
        grille = generation_suivante(grille)
        time.sleep(1)
        
if __name__ == '__main__':
    doctest.testmod()
